/*
  This file is part of the Ayyi Project. http://ayyi.org
  copyright (C) 2004-2013 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#define __ayyi_client_c__
#define __ayyi_private__
#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h>
#include <glib.h>
#include <gmodule.h>

typedef void action;
#include <ayyi/ayyi_typedefs.h>
#include <ayyi/ayyi_types.h>
#include <ayyi/interface.h> //good to remove?
#include <ayyi/ayyi_utils.h>
#include <ayyi/ayyi_shm.h>
#include <ayyi/ayyi_msg.h>
#include <ayyi/ayyi_dbus.h>
#include <ayyi/ayyi_dbus_proxy.h>
#include <ayyi/ayyi_dbus.h>

#include <ayyi/ayyi_client.h>

AyyiService* known_services[2];

#include "services/ardourd.h"
#define plugin_path "/usr/lib/ayyi/:/usr/local/lib/ayyi/"

//FIXME doesnt belong here
extern AyyiItem*       ayyi_song_container_next_item  (AyyiContainer*, void* prev);
extern void            ayyi_song__print_playlists     ();

typedef	AyyiPluginPtr  (*infoFunc)();
typedef void           (*AyyiPluginCallback)(AyyiPluginPtr, GError*, gpointer);

static void          ayyi_client_ping                 (void (*pong)(const char* s));
static void          ayyi_plugin_load                 (const gchar* filename, AyyiPluginCallback, gpointer);
static void          ayyi_client_discover_launch_info ();

enum {
    PLUGIN_TYPE_1 = 1,
    PLUGIN_TYPE_MAX
};
#define PLUGIN_API_VERSION 1


AyyiClient*
ayyi_client_init()
{
	if(ayyi.priv){ gwarn("already initialised!"); return &ayyi; }

	ayyi_log_init();

	ayyi.got_shm = 0;     //before this is set, we must have _all_ required shm pointers.

	ayyi.priv           = g_new0(AyyiClientPrivate, 1);
	ayyi.priv->on_shm   = NULL;
	ayyi.priv->trans_id = 1;
	ayyi.priv->responders = g_hash_table_new(g_str_hash, g_str_equal);

	ayyi_shm_init ();

	known_services[0] = (AyyiService*)&ardourd_song;
dbg(1, "known0=%p=%s", known_services[0], known_services[0]->service);
	known_services[1] = (AyyiService*)&ardourd_mixer;
dbg(1, "known1=%p=%s", known_services[1], known_services[1]->service);

	ayyi.services = g_malloc0(3 * sizeof(gpointer));

	ayyi_client_discover_launch_info();

	return &ayyi;
}


void
ayyi_client_add_services(AyyiService** services)
{
 	ayyi.service = known_services[AYYI_SERVICE_AYYID1];

	if(services){
		ayyi.services = services;
		int i; for(i=0;services[i];i++){
			AyyiService* service = services[i];
			service->priv = g_new0(AyyiServicePrivate, 1);
			service->ping = ayyi_client_ping;

			ayyi_shm_seg_new(service, ((AyyiSongService*)service)->segments[0]);
		}
	}else{
		// default service setup

		ayyi.services[0] = known_services[AYYI_SERVICE_AYYID1];
		ayyi.services[0]->priv = g_new0(AyyiServicePrivate, 1);
		ayyi.services[0]->ping = ayyi_client_ping;

		ayyi.services[1] = known_services[AYYI_SERVICE_MIXER];
		ayyi.services[1]->priv = g_new0(AyyiServicePrivate, 1);
		ayyi.services[1]->ping = ayyi_client_ping;

		// note that addition of shm segments is now handled internally
		ayyi_shm_seg_new(known_services[AYYI_SERVICE_AYYID1], SEG_TYPE_SONG);
		ayyi_shm_seg_new(known_services[AYYI_SERVICE_MIXER], SEG_TYPE_MIXER);
	}
}


gboolean
ayyi_client_connect(AyyiService* service, AyyiHandler2 callback, gpointer user_data, GError** Xerror)
{
	PF;
	gboolean ok = true;

	void temp_copy_shared_proxy()
	{
		if(!known_services[AYYI_SERVICE_MIXER]->priv->proxy) known_services[AYYI_SERVICE_MIXER]->priv->proxy = known_services[AYYI_SERVICE_AYYID1]->priv->proxy;
	}

	void on_connected(GError* error, gpointer user_data)
	{
		dbg(1, "connected!");
		temp_copy_shared_proxy();
		HandlerData* c = user_data;
		call((AyyiHandler2)c->callback, error, c->user_data);
		g_free(c);
	}

	if(service == ayyi.services[0] || service != known_services[AYYI_SERVICE_MIXER]){ // dbus connection is temporarily shared

		HandlerData* c = g_new0(HandlerData, 1);
		c->callback = (AyyiHandler)callback;
		c->user_data = user_data;

		ayyi_client__dbus_connect(service, on_connected, c);
//		temp_copy_shared_proxy();
	}else{
		dbg(1, "NOT connecting... service=%p (0=%p)", service, ayyi.services[0]);
		temp_copy_shared_proxy();
		callback(NULL, user_data);
	}
	return ok;
}


static void
ayyi_client_ping(void (*pong)(const char* s))
{
	ayyi_dbus_ping(ayyi.service, pong);
}


void
ayyi_client_load_plugins()
{
	if(!g_module_supported()) g_error("Modules not supported! (%s)", g_module_error());

	typedef struct _c {
		int     i;
		int     found;
		gchar** paths;
		GDir*   dir;
		gchar*  filename;
		void    (*next_dir)(struct _c*);
		void    (*next_plugin)(struct _c*);
	} C;
	C* c = g_new0(C, 1);
	c->i = -1;

	void plugin_on_load(AyyiPluginPtr plugin, GError* error, gpointer user_data)
	{
		C* c = user_data;
		if(error){
			dbg(1, "'%s' failed to load.", c->filename);
			return;
		}

		c->found++;
		ayyi.priv->plugins = g_slist_append(ayyi.priv->plugins, plugin);

		c->next_plugin(c);
	}

	void done(C* c)
	{
		log_print(0, "ayyi plugins loaded: %i.", c->found);

		if(c->filename) g_free(c->filename);
		g_strfreev(c->paths);
		g_free(c);
	}

	void next_plugin(C* c)
	{
		const gchar* filename = g_dir_read_name(c->dir);
		if (filename) {
			dbg(1, "testing %s...", filename);
			char* path = c->paths[c->i];
			gchar* filepath = g_build_filename(path, filename, NULL);
			// filter files with correct library suffix
			if(!strncmp(G_MODULE_SUFFIX, filename + strlen(filename) - strlen(G_MODULE_SUFFIX), strlen(G_MODULE_SUFFIX))) {
				// If we find one, try to load plugin info and if this was successful try to invoke the specific plugin
				// type loader. If the second loading went well add the plugin to the plugin list.
				if(c->filename) g_free(c->filename);
				c->filename = g_strdup(filename);
				ayyi_plugin_load(filepath, plugin_on_load, c);
			} else {
				dbg(2, "-> no library suffix");
				c->next_plugin(c);
			}
			g_free(filepath);
		} else {
			c->next_dir(c);
		}
	}
	c->next_plugin = next_plugin;

	void next_dir(C* c)
	{
		char* path = c->paths[++c->i];
		dbg(2, "%i: path=%s", c->i, path);
		if(!path){
			done(c);
			return;
		}
		if(!g_file_test(path, G_FILE_TEST_EXISTS)){
			c->next_dir(c);
			return;
		}

		GError* error = NULL;
		c->dir = g_dir_open(path, 0, &error);
		if (!error) {
			next_plugin(c);
		} else {
			gwarn("dir='%s' failed. %s", path, error->message);
			g_error_free(error);
			error = NULL;
		}
	}
	c->next_dir = next_dir;

	c->paths = g_strsplit(plugin_path, ":", 0);

	next_dir(c);
}


AyyiPluginPtr
ayyi_client_get_plugin(const char* name)
{
	//lookup the plugin pointer for a loaded plugin.

	static int count = -1;
	count++;

	GSList* l = ayyi.priv->plugins;
	for(;l;l=l->next){
		AyyiPluginPtr plugin = l->data;
		if(!count) dbg(2, "  %s", plugin->name);
		if(!strcmp(name, plugin->name)) return plugin;
	}
	return NULL;
}


void
ayyi_plugin_load(const gchar* filepath, AyyiPluginCallback callback, gpointer user_data)
{
#warning ayyi_plugin_load will not return anything as it is now async
	AyyiPluginPtr plugin = NULL;
	gboolean success = FALSE;

#if GLIB_CHECK_VERSION(2,3,3)
	GModule* handle = g_module_open(filepath, G_MODULE_BIND_LOCAL);
#else
	GModule* handle = g_module_open(filepath, 0);
#endif

	if(!handle) {
		gwarn("cannot open %s (%s)!", filepath, g_module_error());
		return;
	}

	infoFunc plugin_get_info;
	if(g_module_symbol(handle, "plugin_get_info", (void*)&plugin_get_info)) {
		// load generic plugin info
		if(NULL != (plugin = (*plugin_get_info)())) {
			// check plugin version
			if(PLUGIN_API_VERSION != plugin->api_version){
				dbg(0, "API version mismatch: \"%s\" (%s, type=%d) has version %d should be %d", plugin->name, filepath, plugin->type, plugin->api_version, PLUGIN_API_VERSION);
			}

			// try to load specific plugin type symbols
			switch(plugin->type) {
				default:
					if(plugin->type >= PLUGIN_TYPE_MAX) {
						dbg(0, "Unknown or unsupported plugin type: %s (%s, type=%d)", plugin->name, filepath, plugin->type);
					} else {
						dbg(1, "name='%s' %s", plugin->name, plugin->service_name);

						AyyiService pservice = {plugin->service_name, plugin->app_path, plugin->interface, NULL};
						pservice.priv = g_new0(AyyiServicePrivate, 1);
						typedef struct {
							AyyiService        pservice;
							AyyiPluginPtr      plugin;
							AyyiPluginCallback callback;
							gpointer           user_data;
						} C;
						C* c = g_new0(C, 1);
						c->pservice = pservice;
						c->plugin = plugin;

						void on_plugin_connected(GError* error, gpointer user_data)
						{
							C* c = user_data;
							AyyiPluginPtr plugin = c->plugin;
							AyyiService* pservice = &c->pservice;

							if(error){
								switch(error->code){
									case 98742:
										log_print(0, "plugin service not available: %s", plugin->service_name);
										break;
									default:
										warn_gerror("plugin dbus connection failed", &error);
										break;
								}
								call(c->callback, plugin, error, c->user_data);
								return;
							}
							dbg(1, "plugin dbus connection ok.");

							//dbus_register_signals();
							//dbus_server_get_shm(&engine);  //.....err shouldnt we be using this?

							//get shm address:

							plugin->client_data = (struct _spec_shm*)NULL;

							shm_seg* seg = ayyi_shm_seg_new(pservice, AYYI_SEG_TYPE_PLUGIN);

							if(dbus_g_proxy_call(pservice->priv->proxy, "GetShmSingle", &error, G_TYPE_STRING, "", G_TYPE_INVALID, G_TYPE_UINT, &seg->id, G_TYPE_INVALID)){
								ayyi_shm_import(ayyi.service);
								plugin->client_data = seg->address; //TODO does it need to be translated? no, dont think so.
								dbg(2, "client_data=%p offset=%i", plugin->client_data, (uintptr_t)plugin->client_data - (uintptr_t)plugin->client_data);
							} else {
								if(error){
									dbg (0, "GetShm: %s\n", error->message);
									g_error_free(error);
								}
							}

							call(c->callback, plugin, error, c->user_data);

							if(error){
								g_error_free(error);
							}
						}
						ayyi_client__dbus_connect(&pservice, on_plugin_connected, c);
					}
					break;
			}
		}
	} else {
		gwarn("File '%s' is not a valid Ayyi plugin!", filepath);
	}

	if(!success) {
		g_module_close(handle);
		return;
	}
}


static void
ayyi_client_discover_launch_info()
{
	#define ayyi_application_path "/usr/bin/ayyi/clients/:/usr/local/bin/ayyi/clients/"
	gchar** paths = g_strsplit(ayyi_application_path, ":", 0);
	GError* error = NULL;
	char* path;
	int i = 0;
	while((path = paths[i++])){
		if(!g_file_test(path, G_FILE_TEST_EXISTS)) continue;

		dbg(2, "scanning for ayyi applications (%s) ...", path);
		GDir* dir = g_dir_open(path, 0, &error);
		if (!error) {
			gchar* filename;
			while((filename = (gchar*)g_dir_read_name(dir))){
				dbg(2, "testing %s...", filename);
				if(!g_strrstr(filename, ".desktop")) continue;

				gchar* filepath = g_build_filename(path, filename, NULL);

				GError* error = NULL;
				GKeyFile* key_file = g_key_file_new();
				if(g_key_file_load_from_file(key_file, filepath, G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS, &error)){
					gchar* groupname = g_key_file_get_start_group(key_file);
					if(!strcmp(groupname, "Desktop Entry")){
						AyyiLaunchInfo* info = g_new0(AyyiLaunchInfo, 1);
						gchar* keyval;
						if((keyval = g_key_file_get_string(key_file, groupname, "Name", &error))){
							info->name = keyval;
							if((keyval = g_key_file_get_string(key_file, groupname, "Exec", &error))){
								info->exec = keyval;
							}
						}
						if(true){
							ayyi.launch_info = g_list_append(ayyi.launch_info, info);
						} else g_free(info);
					}
				}else{
					printf("unable to load desktop file: %s.\n", error->message);
					g_error_free(error);
					error = NULL;
				}

				g_free(filepath);
			}
			g_dir_close(dir);
		} else {
			gwarn("dir='%s' failed. %s", path, error->message);
			g_error_free(error);
			error = NULL;
		}
	}
}


void
ayyi_client_set_signal_reg(int signals)
{
#ifdef USE_DBUS
	ayyi_dbus__register_signals(signals);
#endif
}


void
ayyi_object_set_bool(AyyiObjType object_type, AyyiIdx obj_idx, int prop, gboolean val, AyyiAction* action)
{
#ifdef USE_DBUS
	dbus_set_prop_bool(action, object_type, prop, obj_idx, val);
#endif
}


gboolean
ayyi_object_is_deleted(AyyiRegionBase* object)
{
	dbg(3, "deleted=%i (%i)", (object->flags & deleted), object->flags);
	return (object->flags & deleted);
}


#define MAKE_RESPONDER_KEY(object_type, op_type, object_idx) g_strdup_printf("%i-%i-%i", object_type, op_type, object_idx);


void
ayyi_client_add_responder(AyyiIdent object, AyyiOpType op_type, AyyiHandler callback, gpointer user_data)
{
	char* key = MAKE_RESPONDER_KEY(object.type, op_type, object.idx);
	dbg(2, "key=%s", key);
	Responder* responder = g_new0(Responder, 1);
	responder->callback = callback;
	responder->user_data = user_data;

	GList* responders = g_list_append(g_hash_table_lookup(ayyi.priv->responders, key), responder);
	g_hash_table_insert(ayyi.priv->responders, key, responders);

#if 0
	{
		GList* handlers = (GList*)ayyi_client_get_responders(object.type, op_type, object.idx);
		dbg(0, "%s n_handlers=%i", ayyi_print_object_type(object.type), g_list_length(handlers));
	}
#endif
}


void
ayyi_client_add_onetime_responder(AyyiIdent id, AyyiOpType op_type, AyyiHandler user_handler, gpointer user_data)
{
	//deletes the handler after first response.
	//TODO does it need a timeout?

	typedef struct
	{
		AyyiOpType  op_type;
		AyyiHandler handler;
		AyyiHandler user_handler;
		void*       user_data;
	} C;

	void onetime_responder_callback(AyyiIdent obj, GError** error, gpointer _data, gpointer _data2)
	{
		C* data = _data;

		if(_data < 1 << 12) data = _data2; // FIXME callback has different signature depending on whether there is a property or not ! see ayyi_handle()

		call(data->user_handler, obj, error, data->user_data);

		if(!ayyi_client_remove_responder(obj.type, data->op_type, obj.idx, data->handler)){
			gwarn("responder not removed.");
		}

		g_free(data);
	}

	C* data = g_new0(C, 1);
	data->op_type = op_type;
	data->handler = onetime_responder_callback;
	data->user_handler = user_handler;
	data->user_data = user_data;

	ayyi_client_add_responder(id, op_type, onetime_responder_callback, data);
}


const GList*
ayyi_client_get_responders(AyyiObjType object_type, AyyiOpType op_type, AyyiIdx object_index)
{
	//the returned list shoulds NOT be freed.

#if 0
	{
		GHashTableIter iter;
		gpointer key, value;
		g_hash_table_iter_init (&iter, ayyi.priv->responders);
		while (g_hash_table_iter_next (&iter, &key, &value)){
			dbg(0, "  %s=%p", (char*)key, value);
		}
	}
#endif

	char* key = MAKE_RESPONDER_KEY(object_type, op_type, object_index);
	dbg(2, "key=%s %s", key, ayyi_print_optype(op_type));
	return g_hash_table_lookup(ayyi.priv->responders, key);
}


gboolean
ayyi_client_remove_responder(AyyiObjType object_type, AyyiOpType op_type, AyyiIdx object_idx, AyyiHandler callback)
{
																													PF0;
	//remove the responder that matches AyyiObjType, AyyiOpType, and callback.
	//-if callback is NULL, all responders that match the other parameters will be removed

	char* key = MAKE_RESPONDER_KEY(object_type, op_type, object_idx);
	GList* responders = g_hash_table_lookup(ayyi.priv->responders, key);
	if(!responders){
		gwarn("not found: %i %i %i", object_type, op_type, object_idx);
		return FALSE;
	}

	gboolean found = FALSE;
	int length = g_list_length(responders);

	GList* l = responders;
	for(;l;l=l->next){
		Responder* responder = l->data;
		if(!callback || responder->callback == callback){
			g_free(responder);
			responders = g_list_remove(responders, responder);
			g_hash_table_replace(ayyi.priv->responders, key, responders);
			found = TRUE;
			break;
		}
	}
	dbg(2, "key=%s found=%i n_handlers=%i", key, found, g_list_length(responders));
	if(g_list_length(responders) > length - 1) gwarn("handler not removed?");

	if(!g_list_length(responders)){
		dbg(2, "no more handlers - removing responder... %s", ayyi_print_object_type(object_type), ayyi_print_optype(op_type));
		g_list_free(responders); //not needed?
		g_hash_table_remove(ayyi.priv->responders, key);
	}
	return TRUE;
}


void
ayyi_client_print_responders()
{
	char* find_first(char* key, char** pos)
	{
		char* s = key + 1;
		char* e = strstr(s, "-");
		*pos = e + 1;
		char* a = g_strndup(key, e - key);
		return a;
	}

	GHashTableIter iter;
	gpointer _key, value;
	g_hash_table_iter_init (&iter, ayyi.priv->responders);
	while (g_hash_table_iter_next (&iter, &_key, &value)){
		char* key = _key;
		char* pos;
		char* a = find_first(key, &pos);
		int object_type = atoi(a);
		dbg(0, "  %s %s", (char*)key, object_type > -1 ? ayyi_print_object_type(object_type) : "ALL");

		char* b = find_first(pos, &pos);
		//AyyiOp op_type = atoi(b);

		g_free(a);
		g_free(b);
	}
}


void
ayyi_discover_clients()
{
	PF;
}


AyyiIdx
ayyi_ident_get_idx(AyyiIdent ident)
{
	return ident.idx;
}


AyyiObjType
ayyi_ident_get_type(AyyiIdent ident)
{
	return ident.type;
}


void
ayyi_launch(AyyiLaunchInfo* info, gchar** argv)
{
	//argv - use NULL to launch the app without extra arguments.

	g_return_if_fail(info);
	dbg(0, "launching application: %s", info->name);

	int len = argv ? g_strv_length(argv) : 0;
	gchar** v = g_new0(gchar*, len + 2);
	v[0] = info->exec;

	if(argv){
		int i; for(i=0;i<len;i++){
			v[i+1] = argv[i];
		}
		dbg(0, "'%s'", g_strjoinv(" ", v));
	}else{
		//gchar args[2][256] = {"", ""};
		//argv = (char**)args;
		//argv[0] = info->exec;
		//v[1] = NULL;
	}

	GError* error = NULL;
	if(!g_spawn_async(NULL, v, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, &error)){
		log_print(LOG_FAIL, "launch %s", info->name);
		GERR_WARN;
	}
	else log_print(0, "launched: %s", v[0]);

	g_free(v);
}


void
ayyi_launch_by_name(const char* name, gchar** argv)
{
	GList* l = ayyi.launch_info;
	for(;l;l=l->next){
		AyyiLaunchInfo* info = l->data;
		if(!strcmp(info->name, name)){
			ayyi_launch(info, argv);
			return;
		}
	}
	gwarn("not found");
}


gboolean
ayyi_verify_playlists()
{
	PF2;

	extern gboolean ayyi_song_container_verify(AyyiContainer*); //temp
	AyyiContainer* container = &((AyyiSongService*)ayyi.service)->song->playlists;
	if(!ayyi_song_container_verify(container)) return false;

	//TODO shm specific stuff not supposed to be here
	#define ayyi_song__playlist_next(A) (AyyiPlaylist*)ayyi_song_container_next_item(&((AyyiSongService*)ayyi.service)->song->playlists, A)
	#define ayyi_song__audio_track_next(A) (AyyiTrack*)ayyi_song_container_next_item(&ayyi.service->song->audio_tracks, A)
	#define ayyi_song__midi_track_next(A) (AyyiMidiTrack*)ayyi_song_container_next_item(&ayyi.service->song->midi_tracks, A)
	extern AyyiTrack* ayyi_song__get_track_by_playlist (AyyiPlaylist*);

	AyyiPlaylist* playlist = NULL;
	while((playlist = ayyi_song__playlist_next(playlist))){
		dbg(2, "  %i: %s", playlist->shm_idx, playlist->name);
		if(playlist->track == -1){ gwarn("unused playlist: %i: %s", playlist->shm_idx, playlist->name); continue; }
		AyyiTrack* track = ayyi_song__get_track_by_playlist(playlist);
		if(!track){
			dbg(0, "failed for playlist: %s (has invalid track: %i)", playlist->name, playlist->track);
			return false;
		}

		//look for dupes:
		AyyiPlaylist* pl2 = NULL;
		int n = 0;
		while((pl2 = ayyi_song__playlist_next(pl2))){
			if(pl2->track == playlist->track) n++;
		}
		if(n != 1){
			dbg(0, "duplicate found. playlist=%i %s track=%i", playlist->shm_idx, playlist->name, playlist->track);
			ayyi_song__print_playlists();
			return false;
		}
	}
	return true;
}


const char*
ayyi_print_object_type (AyyiObjType object_type)
{

	static char unk[32];
  
	switch (object_type){
#define CASE(x) case AYYI_OBJECT_##x: return "AYYI_OBJECT_"#x
		CASE (EMPTY);
		CASE (TRACK);
		CASE (AUDIO_TRACK);
		CASE (MIDI_TRACK);
		CASE (CHAN);
		CASE (AUX);
		CASE (PART);
		CASE (AUDIO_PART);
		CASE (MIDI_PART);
		CASE (EVENT);
		CASE (RAW);
		CASE (STRING);
		CASE (ROUTE);
		CASE (FILE);
		CASE (LIST);
		CASE (MIDI_NOTE);
		CASE (SONG);
		CASE (TRANSPORT);
		CASE (AUTO);
		CASE (UNSUPPORTED);
		CASE (ALL);
#undef CASE
		default:
			snprintf (unk, 31, "UNKNOWN OBJECT TYPE (%d)", object_type);
			return unk;
	}
	return NULL;
}


const char*
ayyi_print_optype (AyyiOp op)
{
	static char* ops[] = {"", "AYYI_NEW", "AYYI_DEL", "AYYI_GET", "AYYI_SET", "AYYI_UNDO", "BAD OPTYPE"};
	return ops[MIN(op, 6)];
}


const char*
ayyi_print_prop_type (uint32_t prop_type)
{
	static char unk[32];

	switch (prop_type){
#define CASE(x) case AYYI_##x: return "AYYI_"#x
		CASE (NO_PROP);
		CASE (NAME);
		CASE (STIME);
		CASE (LENGTH);
		CASE (LEVEL);
		CASE (PAN);
		CASE (PAN_ENABLE);
		CASE (DELAY);
		CASE (HEIGHT);
		CASE (COLOUR);
		CASE (END);
		CASE (TRACK);
		CASE (MUTE);
		CASE (ARM);
		CASE (SOLO);
		CASE (SDEF);
		CASE (INSET);
		CASE (FADEIN);
		CASE (FADEOUT);
		CASE (INPUT);
		CASE (OUTPUT);
		CASE (PREPOST);
		CASE (SPLIT);
		CASE (PLUGIN_SEL);
		CASE (PLUGIN_BYPASS);
		CASE (PB_LEVEL);
		CASE (PB_PAN);
		CASE (PB_DELAY);
		CASE (TRANSPORT_PLAY);
		CASE (TRANSPORT_STOP);
		CASE (TRANSPORT_REW);
		CASE (TRANSPORT_FF);
		CASE (TRANSPORT_REC);
		CASE (TRANSPORT_LOCATE);
		CASE (TRANSPORT_CYCLE);
		CASE (TRANSPORT_LOCATOR);
		CASE (AUTO_PT);
		CASE (ADD_POINT);
		CASE (DEL_POINT);
		CASE (TEMPO);
		CASE (HISTORY);
		CASE (LOAD_SONG);
		CASE (SAVE);
		CASE (NEW_SONG);
#undef CASE
		default:
			snprintf (unk, 31, "UNKNOWN PROPERTY (%d)", prop_type);
	}
	return unk;
}


const char*
ayyi_print_media_type (uint32_t type)
{
	static char* types[] = {"[Media type not set]", "AUDIO", "MIDI"};
	if(G_N_ELEMENTS(types) != AYYI_MIDI + 1) gerr("!!");
	return types[type];
}


