/*
  This file is part of AyyiGtk. http://ayyi.org
  copyright (C) 2004-2012 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <math.h>
#include <gtk/gtk.h>
#include "common.h"
#include <ayyi/ayyi_list.h>
#include <ayyi/ayyi_shm.h>
#include <ayyi/ayyi_client.h>
#include <ayyi/interface.h>
#include <ayyi/ayyi_song.h>
#include <ayyi/ayyi_mixer.h>
#include <ayyi/ayyi_action.h>
#include "ayyi/ayyi_dbus_proxy.h"

#include "model/am_time.h"
#include "model/am_message.h"
#include "model/am_song.h"
#include "model/am_curve.h"
#include "model/am_plugin.h"
#include "model/am_track_peer.h"
#include "model/am_automation.h"

extern gboolean is_mixer_shm_local(void*);

//gui fn (TODO need to store automation positions in samples)
#warning TODO need to store automation positions in samples not pixels
#define PX_PER_BEAT 8.0
static double
_samples2px_nz(uint64_t samples)
{
	float beats_per_second = ((AyyiSongService*)ayyi.service)->song->bpm / 60.0;

	float secs = (float)samples / (float)song->sample_rate;
	float beats = 4 * secs / beats_per_second; //why *4 ?

	double px = beats * PX_PER_BEAT;

	return px;
}


static uint32_t
_px2samples_nz(double px_nonzoomed)
{
  //converts an arrange window pixel value to a sample value.

  double beats = px_nonzoomed / PX_PER_BEAT;

  //check:
  //double error = samples2px(ayyi_beats2samples_float(beats)) - px_zoomed;
  //printf(" error=%.1f\n", error);

  dbg (3, "%.2f --> %.2f --> %i", px_nonzoomed, beats, ayyi_beats2samples_float(beats));
  return ayyi_beats2samples_float(beats);
}


static bool am_automation_verify_path(Curve*);


gboolean
am_automation_sync()
{
	if(!((AyyiSongService*)ayyi.service)->amixer) return false;

	PF;
	int t; for(t=0;t<am_track_list_count(song->tracks);t++){
		AMTrack* trk = song->tracks->track[t];
		if(!TRK_IS_AUDIO(trk)) continue;

		AyyiChannel* ch = ayyi_mixer__channel_at_quiet(trk->ident.idx);
		int auto_type; for(auto_type = 0; auto_type < AUTO_MAX; auto_type++){
			AyyiContainer* container = &ch->automation[auto_type];
			dbg(2, "t=%i type=%i...", t, auto_type);
			if(!ch->automation[auto_type].block[0]){
				dbg (2, "track %i has no type%i automation events. abp=%p &=%p=%p", t, auto_type, ((BPath**)&trk->bezier)[auto_type], &trk->bezier.pan, &((BPath**)&trk->bezier)[auto_type]);
				continue;
			}
			//we cant verify the container as it is non-standard - items have no idx.
			if(false && !ayyi_client_container_verify(container, translate_mixer_address)) continue;
			int count = ayyi_mixer_container_count_items(&ch->automation[auto_type]);
			if(count < 2) { gwarn ("curve is too short - must have at least 2 points... size=%i type=%i", count, auto_type); continue; }//not really sure if this is an error. We should perhaps add it to the curve, then not draw it.
			dbg(2, "t=%i type=%i: count=%i", t, auto_type, count);
			if(count > 10) ayyi_container_print(&ch->automation[auto_type]);

			int event_num = 0;
			//TODO move this curve stuff to curve.c
			BPath* abp = NULL;

			Curve* curve = ((Curve**)&trk->bezier)[auto_type];

			double prev = 0;
			AyyiContainer* events = &ch->automation[auto_type];
			shm_event* shared = NULL;
			int i = 0;
			while((shared = ayyi_mixer__container_next_item(events, shared))){
				if(!is_mixer_shm_local(shared)){ gwarn("event %i not in mixer shm", i); break; }
				//if(!is_song_shm_local_quiet(shared)){ gwarn("event %i: not song shm", i); break; }
				if(shared->when < 0){ gwarn ("bad data! automation event time cannot be negative! seg=%i type=%i.", i, auto_type); return false; }
				if(shared->when < prev) gwarn ("automation list not sorted properly seg=%i type=%i.", i, auto_type);
				if(shared->value > 10.0){ gwarn ("automation value out of range. %.2f. Ignoring...", shared->value); continue; }

				abp = curve_append(curve);
				g_return_val_if_fail(abp, false);

				abp->code = event_num ? LINETO : MOVETO_OPEN;
				abp->x1 =   0.0;   abp->y1 =   0.0;  
				abp->x2 =   0.0;   abp->y2 =   0.0;
				abp->x3 =  _samples2px_nz(shared->when);  abp->y3 = ayyi_automation_xlate_in(shared->value);

				if(_debug_ > 2) printf("  %10.2f: %4.2f-->%.2f\n", shared->when, shared->value, abp->y3);

				prev = shared->when;
				event_num++; 
				i++;
			}
			if(abp){
				//gnome canvas doesnt show the end point, so we have to add an extra one.
				BPath* end = curve_append(curve);
				dbg (2, "e=%i x=%.2f", event_num, abp->x3);
				end->code = BP_END;
				end->x3 = abp->x3; end->y3 = abp->y3;
			}
dbg(2, "curvelen=%i", curve_get_length(curve));

			am_song__emit("curve-change", curve);
		}

		//automation controls:
		auto_type = 2;
		if(!ch->automation_list) dbg(2, "t=%i: no plugin automation.", t);
		//dbg(0, "len=%i", ayyi_list__length(ch->automation_list));
		/*
		AyyiList* l; for(l=ayyi_list__first(ch->automation_list); l; l=ayyi_list__next(l)){
			AyyiContainer* events = ayyi_song_translate_address(l->data);
		*/
		AyyiList* l = ch->automation_list;
		AyyiContainer* events = ayyi_list__first_data(&l);
//#ifdef USE_GNOMECANVAS
		for(;l;(events = ayyi_list__next_data(&l))){
			if(!ayyi_client_container_verify(events, translate_mixer_address)) continue;
			dbg(0, "t=%i type=%i control_name=%s n_events=%i", t, auto_type, l->name, ayyi_mixer_container_count_items(events));
			//TODO the automationlist should use ->id to tell us which plugin it belongs to.
			//ayyi_list->type is not being used, so we can use that.
			//-we need to know the plugin, the insert num, and the control num. 

			//for now we take the first plugin:
			AMPlugin* plugin = am_plugin__get_by_idx(0);

			Curve* curve = curve_new(trk, auto_type, plugin);
			trk->controls = g_list_append(trk->controls, curve);
			auto_type++;

			shm_event* event = NULL;
			BPath* abp = NULL;
			int event_num = 0;
			double prev = 0;
			int i = 0;
			while((event = ayyi_container_next_item(events, (AyyiItem*)event, translate_mixer_address))){
				if(event->when < prev) gwarn ("automation list not sorted properly seg=%i type=%i.", i, auto_type);
				if(event->value > 10.0){ gwarn ("automation value out of range. %.2f. Ignoring...", event->value); continue; }

				abp = curve_append(curve);
				g_return_val_if_fail(abp, false);

				abp->code = event_num ? LINETO : MOVETO_OPEN;
				abp->x1 =   0.0;   abp->y1 =   0.0;  
				abp->x2 =   0.0;   abp->y2 =   0.0;
				abp->x3 =  _samples2px_nz(event->when); abp->y3 = ayyi_automation_xlate_in(event->value);

				if(_debug_ > 2) printf("  %10.2f: %4.2f-->%.2f\n", event->when, event->value, abp->y3);

				prev = event->when;
				event_num++; 
				i++;
			}
		}
//#endif
	}
	return TRUE;
}


void
am_automation_point_add(Curve* curve, int p, AyyiHandler callback, gpointer user_data)
{
	// ask the server to add a point to the given automation curve
	//
	// the new point should already have been added to the local Curve.
	// (TODO this is not very versatile and should change)

	//			automation_print_bezier_path(curve);
		if(!am_automation_verify_path(curve)) gerr("bad path!");
	PF0;

	/*
	void am_automation_point_add_cb(AyyiAction* a)
	{
		PF0;
		HandlerData* d = a->app_data;
		if(d) call(d->callback, AYYI_NULL_IDENT, &a->ret.error, d->user_data);
	}
	*/

	BPath* bpath = (BPath*)curve->path;
	BPath* bp  = &bpath[p];

	int plugin_slot = 0; //FIXME

	AyyiAction* a = ayyi_action_new("automation point add");
	//a->callback   = am_automation_point_add_cb;
	a->callback   = ayyi_handler_simple;
	a->app_data   = ayyi_handler_data_new(callback, user_data);
#ifdef USE_DBUS
	dbus_set_prop_f_pair(a, AYYI_OBJECT_AUTO, AYYI_ADD_POINT, ayyi_index_array_new(curve->track->ident.idx, curve->auto_type, plugin_slot), _px2samples_nz(bp->x3), ayyi_automation_xlate_out(bp->y3));
#endif
}


void
am_automation_point_remove(Curve* curve, int p, AyyiHandler callback, gpointer user_data)
{
	PF0;
	void am_automation_point_remove_cb(AyyiAction* a)
	{
		//this is only needed temporarily while a signal is not emitted from server for automation point remove

		HandlerData* d = a->app_data;
		if(d) call(d->callback, AYYI_NULL_IDENT, &a->ret.error, d->user_data);
	}

	int len = curve_get_length(curve);
	curve_del(curve, p); //tmp!
	dbg(0, "len: %i --> %i", len, curve_get_length(curve));

	AyyiAction* a = ayyi_action_new("automation point remove");
	a->callback     = am_automation_point_remove_cb;
	/*
	a->callback   = ayyi_handler_simple;
	*/
	a->app_data   = ayyi_handler_data_new(callback, user_data);

	dbus_set_prop_f_pair(a, AYYI_OBJECT_AUTO, AYYI_DEL_POINT, ayyi_index_array_new(curve->track->ident.idx, curve->auto_type, p), 0.0, 0.0);
}


void
am_automation_point_change(Curve* curve, int p, AyyiHandler callback, gpointer user_data)
{
	void am_automation_point_change_cb(AyyiAction* action)
	{
	}

	BPath* bpath = curve->path;
	BPath* bp    = &bpath[p];
	g_return_if_fail(bp);

	AyyiAction* a   = ayyi_action_new("automation update");
	a->callback     = am_automation_point_change_cb;
	a->op           = AYYI_SET;
	a->obj.type     = AYYI_OBJECT_AUTO;
	a->prop         = AYYI_AUTO_PT;
	a->obj_idx.idx1 = curve->track->ident.idx;
	a->obj_idx.idx2 = curve->auto_type;
	a->obj_idx.idx3 = p;
	a->d_val        = _px2samples_nz(bp->x3);
	a->val2.d       = ayyi_automation_xlate_out(bp->y3);

	dbg (0, "p=%i x=%.2f y=%.2f-->%.2f", p, a->d_val, bp->y3, a->val2.d);

	action_execute(a);
}


double
ayyi_automation_xlate_in(double gain)
{
	//converts from ardour gain values to normalised canvas values

	//server-interface y values are not logarithmic and go from 0.0 to 2.0 (1.0 is unity gain).
	//internal y values use log scale and go from 100.0 to 0.0.

	//TODO do other automation types use the same scale?

	/*
	if(in > 1.0){
		//map 1.0 -> 4.0 to 70(?)-100
		//y = 70.0 + (in - 1.0) * (30 / 3);
		y = _map(in, 1.0, 4.0, 75.0, 100.0);
	}else{
		y = 65 * (in - 0.1);
	}
	*/

	//from ardour:
	double y = (gain == 0)
		? 0
		: pow((6.0*log(gain)/log(2.0)+192.0)/198.0, 8.0);

	dbg (2, "y=%.2f-->%.2f", gain, y);

	return 100 - 100 * y;
}


double
ayyi_automation_xlate_out(double val)
{
	// @val:    in range 100 to 0. Logarithmic.
	// @return: normally 0 to 2.0 giving us max 6dB gain.

	if(val < 0.0 || val > 100.0){ gerr ("bad value: %.2f", val); return 0.0; }

	if(val == 0.0) return 2.0; //check!! shouldnt be needed, but...

	val = (100.0 - val) / 100; //normalise

	if (val == 0.0) val = 0.0;
	else            val = pow (2.0,(sqrt(sqrt(sqrt(val)))*198.0-192.0)/6.0);

	return val;
}


void
am_automation_curve_changed(Curve* curve)
{
	//trigger gui updates from a changed source BPath.
	dbg(0, "curve='%s'", curve->name);

if(song->loaded) gwarn("emitting...");
	am_song__emit ("curve-change", curve);
}


static bool
am_automation_verify_path(Curve* curve)
{
	size_t len = curve_get_length(curve);
	BPath* bezier = (BPath*)curve->path;
	BPath* abp = &bezier[len -1];
	if(abp->code != BP_END) gwarn("end should be BP_END ?");
	return true;
}


void
am_automation_print_path_code(int i, char* s)
{
  switch(i){
    case MOVETO_OPEN:
      strcpy(s, "ART_MOVETO_OPEN");
      break;
    case LINETO:
      strcpy(s, "ART_LINETO");
      break;
    case BP_END:
      strcpy(s, "ART_END");
      break;
    default:
      sprintf(s, "%i", i);
      break;
  }
}


void
am_automation_print_bezier_path(Curve* curve)
{
	BPath* abp = curve->path;

	if(!abp) dbg (0, "path is empty.");

	int i; for(i=0;i<curve_get_length(curve);i++){
		BPath* p = &abp[i];
		if(!p) break;

		char code[64]; am_automation_print_path_code(p->code, code);

		printf(" %02i: %15s %.3f %.2f\n", i, code, p->x3, p->y3);
	}
	printf("--------------------\n");
}


