#ifndef __am_channel_h__
#define __am_channel_h__

#include "model/am_collection.h"

typedef enum{
  CHAN_INPUT = 1,
  CHAN_BUS,
  CHAN_GROUP,  //vca type control strip
  CHAN_MASTER
} ChannelType;

#define MAX_CHS 32
#define CH_IS_MASTER(C) (C && C->type == CHAN_MASTER)
#define ASSERT_CH_NUM(A)  if(A < 0 || A > MAX_CHS){ errprintf2(__func__, "%s(): chan num of out range (%i).\n", A); return NULL; }

#define channel_iter_init(ITER) model_collection_iter_init((ModelCollection*)song->channels, ITER)
#define channel_next(ITER) model_collection_iter_next((ModelCollection*)song->channels, ITER)
#define am_channels ((ModelCollection*)song->channels)
#define am_channels__emit(...) ({ if(song->loaded) g_signal_emit_by_name (song->channels, __VA_ARGS__); })

// -channels exist only as members of the GArray song->channels.
struct _channel {
	ChannelType   type;
	int           nchans;                          // only stereo or mono supported.
	AyyiIdx       core_index;

	GtkObject*    aux_level[AYYI_AUX_PER_CHANNEL]; // GtkAdjustment*
	GtkObject*    meter_level;                     // GtkAdjustment*

	GtkTreeModel* plugin_list;
};


Channel*    channels__get_master         ();
Channel*    channels__get_by_index       (AyyiIdx);

const char* am_channel__get_name         (Channel*);
double      am_channel__fader_level      (Channel*);
double      am_channel__pan_value        (Channel*, bool* has_pan);
bool        am_channel__is_muted         (Channel*);
bool        am_channel__is_solod         (Channel*);
bool        am_channel__is_sdef          (Channel*);
bool        am_channel__is_armed         (Channel*);
int         am_channel__get_input_idx    (Channel*);
void        am_channel__get_input_string (Channel*, char*);
void        am_channel__get_output_string(Channel*, char*);
void        am_channel__set_enable_pan   (Channel*, gboolean enable, AyyiHandler, gpointer);

double      am_channel__get_aux_level    (Channel*, int aux_num);
void        am_channel__set_aux_level    (Channel*, int aux_num, float, AyyiHandler, gpointer);
void        am_channel__set_aux_output   (Channel*, int aux_num, int connection, AyyiHandler, gpointer);

AyyiConnection* am_channel__get_connected_to (Channel*);

bool        am_channel__is_type          (Channel*, gpointer);

//private

#ifdef __am_song_c__
AMArray*    channels__init               (int n_channels);
void        channels__load               ();
void        channels__append_new         (TrackNum);
void        channels__remove_channel     (AMTrack*);
#endif

#endif //__am_channel_h__
