/*
  This file is part of AyyiGtk. http://ayyi.org
  copyright (C) 2004-2012 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <common.h>
#include "math.h"
#include <ayyi/ayyi_time.h>
#include <ayyi/ayyi_client.h>
#include <ayyi/ayyi_mixer.h>
#include <ayyi/ayyi_msg.h>
#include <ayyi/ayyi_action.h>
#include <ayyi/ayyi_song.h>
#include <ayyi/ayyi_shm.h>
#include <model/am_collection.h>
#include <model/am_channel.h>
#include <model/am_message.h>
#include <model/am_mixer.h>
#include <model/am_song.h>


void
am__mixer_sync()
{
	if(!((AyyiSongService*)ayyi.service)->amixer) return;

	int n_mixer_tracks = ayyi_mixer_container_count_items(&((AyyiSongService*)ayyi.service)->amixer->tracks);
	int n_song_tracks = ayyi_song_container_count_items(&((AyyiSongService*)ayyi.service)->song->audio_tracks);
	if(n_mixer_tracks != n_song_tracks){ gerr("shm error. song and mixer out of sync. song=%i mixer=%i", n_song_tracks, n_mixer_tracks); }

	int i = 0;
	AyyiChannel* track = NULL;
	while((track = ayyi_mixer__channel_next(track))){
		dbg (2, "%i: track=%p", i, track->name);

		int n_aux = 0;
		int a; for(a=0;a<AYYI_AUX_PER_CHANNEL;a++){
			AyyiAux* aux = ayyi_mixer__aux_get(track, a);
			if(aux) n_aux++;
		}
		dbg(3, "%i: n_aux=%i", i, n_aux);

		i++;
	}
	if(_debug_ > 1) am__mixer_print();
}


void
am__mixer_periodic_update(GObject* _song, gpointer user_data)
{
	AMIter iter;
	channel_iter_init(&iter);
	Channel* channel;
	while((channel = channel_next(&iter))){
		int aux_num; for(aux_num=0;aux_num<AYYI_AUX_PER_CHANNEL;aux_num++){
			AyyiAux* aux = ayyi_mixer__aux_get_(channel->core_index, aux_num);
			if(aux){
				double aux_level_db = am_channel__get_aux_level(channel, aux_num);
				//if(c==0 && aux_num==0 && !(jjj++ % 100)) dbg(0, "gain=%.2f dB=%.2f", aux_level, gain2db(aux_level));
				GtkAdjustment* adj = (GtkAdjustment*)channel->aux_level[aux_num];
				if(aux_level_db != adj->value){
					if(aux_level_db > gtk_adjustment_get_upper(adj) || isnan(aux_level_db)){
						gwarn("out of range: %.2fdB", aux_level_db);
					}else{
						//if(!aux_num) dbg(0, "setting: %.4f (was %.4f)", aux_level_db, adj->value);
						gtk_adjustment_set_value(adj, aux_level_db);
					}
				}
			}
		}
	}
}


unsigned
am__mixer_channel_next(const Channel* channel)
{
	//returns the pod index for the strip following the one given.

	AyyiChannel* prev = channel ? ayyi_mixer__channel_at_quiet(channel->core_index) : NULL;
	AyyiChannel* next = ayyi_mixer__channel_next(prev);
	return (next ? next->shm_idx : 0);
}


void
am__mixer_show_routing()
{
}


void
am_aux__add(Channel* ch, int aux_num, AyyiHandler callback, gpointer user_data)
{
	g_return_if_fail(ch);

	am_channel__set_aux_output(ch, aux_num, -1, callback, user_data);
}


void
am_aux__remove(Channel* ch, int aux_num, AyyiHandler callback, gpointer user_data)
{
	g_return_if_fail(ch);

	void am_aux__remove_done(AyyiAction* action)
	{
		g_return_if_fail(action);

		HandlerData* d = action->app_data;
		if(d){
dbg(0, "callback=%p", d->callback);
			call(d->callback, (AyyiIdent){action->obj.type, action->obj_idx.idx1}, &action->ret.error, d->user_data);
		}
	}

	AyyiAction* a   = ayyi_action_new("delete aux (track %i, aux %i)", ch->core_index, aux_num);
	a->op           = AYYI_DEL;
	a->obj.type     = AYYI_OBJECT_AUX;
	a->obj_idx.idx1 = (ch->core_index << 4) | aux_num; // warning! index is combination of auxnum and channel
dbg(0, "idx=0x%x callback=%p", a->obj_idx.idx1, callback);
	a->callback     = am_aux__remove_done;
	a->app_data     = ayyi_handler_data_new(callback, user_data);
	action_execute(a);
}


void
am__mixer_print()
{
	if(!((AyyiSongService*)ayyi.service)->amixer) return;
	if(strcmp(((AyyiSongService*)ayyi.service)->amixer->service_type, "Ayyi mixer")){ gwarn("bad shm signature! %s\n", ((AyyiSongService*)ayyi.service)->amixer->service_type); return; }

	int n_tracks = ayyi_mixer__count_channels();
	printf("------------------------------------------------\n");
	dbg(0, "");
	if(n_tracks > 0) printf("                name  lvl\n");
	int count = 0;
	AyyiChannel* trk = NULL;
	while((trk = ayyi_mixer__channel_next(trk))){
		printf("%02i %02i %14s %.2f %p\n", count, trk->shm_idx, trk->name, trk->level, trk);
		count++;
	}
	printf("total=%i\n", count);
	printf("------------------------------------------------\n");
}


