#ifndef __track_h__
#define __track_h__

#define ASSERT_TRACK_NUM(A) if(A < 0 || A > AM_MAX_TRK){ errprintf2(__func__, "%s(): track number out of range (%i)\n", A); return 0; }

#define TRK_IS_AUDIO(A) (A->type == TRK_TYPE_AUDIO)
#define TRK_IS_MIDI(A) (A->type == TRK_TYPE_MIDI)
#define TRK_TYPE_IS_VALID(A) (A > 0 && A <= TRK_TYPE_MIDI)

//#define TRK_GET_AYYI(T) (TRK_IS_AUDIO(T) ? ayyi_song__audio_track_at(T->ident.idx) : (AyyiTrack*)ayyi_song__midi_track_at(T->ident.idx))
#define am_track__get_shared(T) (TRK_IS_AUDIO((T)) \
	? (AyyiItem*)ayyi_song__audio_track_at((T)->ident.idx) \
	: (TRK_IS_MIDI(T) ? (AyyiItem*)ayyi_song__midi_track_at((T)->ident.idx) : (AyyiItem*)ayyi_song__audio_track_at((T)->ident.idx)))

struct _am_track
{
	//a track object that acts as a proxy for the Ayyi track object.
	//-it multiplexes two Ayyi lists: audio tracks and midi tracks.

	AyyiIdent           ident;
	AMTrackType         type;
	TrackNum            track_num;       // this is a self reference. To get the display number, use this number to lookup in the track_order array.
	char*               name;
	gboolean            visible;         // in theory this should be overridable in arr_trk.
#if 0
	int                 colour;          // palette index.
#endif
	GString*            user_notes;
	GtkObject*          meteradj;        // used by Steve Harris's gtkmeter.

	// -playback params: delay, level, etc 

	struct _track_curves {
		Curve*              vol;         // dynamic array holding the complete bezier path data. is _always_ initialised.
		Curve*              pan;
		int                 size : 2;
	} bezier;

	GList*              controls;        // list of Curve*
};


#if 0
struct _engine_cache {
	int track_count;
};
#endif

#ifdef __am_private__
void            am_track__init                  (TrackNum);
TrackNum        am_track__lookup_index          (AMTrack*);
#endif
void            am_track_reset                  (TrackNum);
AMTrack*        am_track__new                   (AMTrackType, TrackNum, AyyiIdx);

void            am_track__change_ch             (AMTrack*, AyyiIdx, AyyiHandler, gpointer);
void            am_track__get_channel_label     (const AMTrack*, char* label);

void            am_track__arm                   (AMTrack*, gboolean arm_state, AyyiHandler, gpointer);
void            am_track__mute                  (AMTrack*, gboolean mute_state, AyyiHandler, gpointer);
void            am_track__solo                  (AMTrack*, gboolean solo_state, AyyiHandler, gpointer);
void            am_track__set_height            (AMTrack*, int height, AyyiHandler, gpointer);

Channel*        am_track__lookup_channel        (AMTrack*);
int             am_track__lookup_channel_num    (AMTrack*);

guchar          am_track__is_armed              (AMTrack*);
gboolean        am_track__is_solod              (AMTrack*);
gboolean        am_track__is_muted              (const AMTrack*);
bool            am_track__is_master             (AMTrack*);

gboolean        am_track__get_connections_string(const AMTrack*, char* label);
gboolean        am_track__get_name              (const AMTrack*, char* name);
void            am_track__set_name              (const AMTrack*, const char* name, AyyiHandler, gpointer);
int             am_track__get_colour            (AMTrack*);
void            am_track__set_colour            (const AMTrack*, int colour, AyyiHandler, gpointer);
void            am_track__set_input             (AMTrack*, AyyiConnection*);
void            am_track__set_output            (AMTrack*, AyyiConnection*, AyyiHandler, gpointer);
const char*     am_track__get_input_name        (AMTrack*);
const char*     am_track__get_output_name       (AMTrack*);
float           am_track__get_meterlevel        (AMTrack*);
void            am_track__get_meterlevel_master (float level[]);

int             am_track__get_input_num         (AMTrack*);
int             am_track__get_output_num        (AMTrack*);
AyyiConnection* am_track__get_output            (AMTrack*);

Curve*          am_track__lookup_curve          (AMTrack*, int);
void            am_track__clear_automation      (AMTrack*, Curve*, AyyiHandler, gpointer);

void            am_track__enable_metering       (AMTrack*);

bool            am_track__is_empty              (AMTrack*, gpointer);
bool            am_track__is_mono               (AMTrack*, gpointer);

gboolean        am_track__is_valid              (const AMTrack*);
#ifdef __am_private__
gboolean        am_track__exists                (AMTrack*);
void           _am_track__sync_name             (AMTrack*);
#endif

const char*     am_track__print_type            (AMTrack*);

#endif //__track_h__
