/*
  This file is part of AyyiGtk. http://ayyi.org
  copyright (C) 2004-2012 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#define __am_track_c__
#define __am_track__peer_c__
#define __am_private__
#include <model/common.h>
#include <gdk/gdkkeysyms.h>
#include <ayyi/ayyi_list.h>
#include <ayyi/ayyi_shm.h>
#include <ayyi/ayyi_client.h>
#include <ayyi/ayyi_song.h>
#include <ayyi/ayyi_msg.h>
#include <ayyi/ayyi_action.h>

#include <model/am_song.h>
#include <model/am_curve.h>
#include <model/am_channel.h>
#include <model/am_track_peer.h>

#define observer_track_change(A, B, C) observer.tracks_change(A, A, B, C)

extern gboolean offline;
extern int debug;

extern void      am_track__rebuild_pointers       ();

void            _am_track_manager__remove_track   (AMTrack*);


void
am_tracks_init()
{
	//should only ever be called once.
	if(track_pool) gwarn("already initialised! use am_tracks_reset() instead.");

	track_pool = g_malloc0(sizeof(AMTrack) * AM_MAX_TRK);

	int t;
	for(t=0;t<AM_MAX_TRK;t++){
#if 0
		song->tracks->track[t] = track_pool_get_item(t);
#endif
		am_track__init(t);
	}
}


void
am_tracks_reset()
{
	g_return_if_fail(track_pool);

	int t;
	for(t=0;t<AM_MAX_TRK;t++){
		song->tracks->track[t] = track_pool_get_item(t);
		am_track_reset(t);
	}
}


void
_am_track_manager__remove_track(AMTrack* trk)
{
	//remove a trk from the client. Can be used following notification of a track delete from the server.
	PF2;

	trk->type = TRK_TYPE_NULL;

	am_track__rebuild_pointers();

	GList* trks = g_list_append(NULL, trk);
	g_signal_emit_by_name (song, "tracks-delete", trks);
	//g_list_free(trks); -- where is freed?
}


TrackNum
_am_track_manager__find_unused()
{
  //returns an internal track index for the first unused track.

  unsigned t;
  for(t=0;t<AM_MAX_TRK;t++){
    dbg (2, "t=%i track=%p type=%i", t, song->tracks->track[t], song->tracks->track[t] ? song->tracks->track[t]->type : 0);
    if(!song->tracks->track[t] || song->tracks->track[t]->type == TRK_TYPE_NULL) return t;
  }
  return -1;
}


#if 0 //moved
gboolean
am_track_manager__verify_tracks()
{
	return true;
}


#endif


