/*

Test client for Ayyi Plugin loading
-----------------------------------

*/
#include "config.h"
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <getopt.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <signal.h>
#include <glib.h>
#include <jack/jack.h>
#include "model/ayyi_model.h"
#include "test/am_common.h"
#ifdef USE_SPECTROGRAM
#include "spectrogram/client.h"
#endif

static void app_on_shm();

char current_test_name[64];

TestFn test_one;

gpointer tests[] = {
	test_one,
};


int
main (int argc, char* argv[])
{
	am_test_init(current_tests = tests, G_N_ELEMENTS(tests));
	ayyi.log.to_stdout = true;

	GError* error = NULL;
	if ((app.dbus = am__connect_messaging(ayyi.service, NULL, NULL, &error))) {
		ayyi_shm_seg_new(ayyi.service, SEG_TYPE_SONG);
		am__get_shm(ayyi.service, app_on_shm, NULL);

		g_main_loop_run (g_main_loop_new (NULL, 0));
	}

	printf("failed to connect to Ayyi server\n");
	return EXIT_FAILURE;
}


void
test_one()
{
	START_TEST;

#ifdef USE_SPECTROGRAM
	AyyiPluginPtr plugin;
	if((plugin = ayyi_client_get_plugin("Ayyi Spectrogram Plugin"))){
		SpectrogramSymbols* spectrogram = plugin->symbols;
		//spectrogram->on_expose(NULL, NULL, NULL, plugin);
		dbg(0, "hello=%s", spectrogram->get_hello());
		dbg(0, "meter_level=%.2f", spectrogram->get_meterlevel(plugin));
		assert(!strcmp("hello", spectrogram->get_hello()), "unexpected signature");
	}
	else FAIL_TEST("couldnt get plugin");
#else
	assert(false, "USE_SPECTROGRAM not defined");
#endif

	FINISH_TEST;
}


static void
app_on_shm()
{
	static bool done = false;
	if(done) return;
	done = true;

	if(((AyyiSongService*)ayyi.service)->song->version != AYYI_SHM_VERSION){
		gwarn ("wrong shm version: %i - should be %i", ((AyyiSongService*)ayyi.service)->song->version, AYYI_SHM_VERSION);
		return;
	}

	ayyi.got_shm = 1;

	log_print(LOG_OK, "connected to server");

	//synchronise local data cache.
	am_song__load();

	ayyi_client_load_plugins();

	next_test();
}


