/*

tests survival of a server restart

usage: once the program has connected, stop then restart the server.

*/
#define __ayyi_private__
#include "config.h"
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <getopt.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <signal.h>
#include <glib.h>
#include <jack/jack.h>
#include "model/ayyi_model.h"
#include "test/am_common.h"

void              app_on_all_shm();

static void       test__wait_for_server_restart();

#define STOP false;
char       current_test_name[64];
gpointer   tests[] = {
	test__wait_for_server_restart,
};

#define assert(A, B, ...) \
	{bool __ok_ = (bool)A; \
	{if(!__ok_) gerr(B, ##__VA_ARGS__); } \
	{if(!__ok_) FAIL_TEST("assertion failed") }}

#define assert_and_stop(A, B, ...) \
	{bool __ok_ = (bool)A; \
	{if(!__ok_) gerr(B, ##__VA_ARGS__); } \
	{if(!__ok_) FAIL_TEST_TIMER("assertion failed") }}


#include "modules.c"


int
main (int argc, char *argv[])
{
	am_test_init(current_tests = tests, G_N_ELEMENTS(tests));
	ayyi.log.to_stdout = true;

	app.dbus = true; //TODO
	am__connect_all(app_on_all_shm, NULL);
	g_main_loop_run (main_loop = g_main_loop_new (NULL, 0));

	return EXIT_SUCCESS;
}


static void
test__wait_for_server_restart()
{
	START_TEST;

	void __on_song_new(AyyiIdent obj, GError** errror, gpointer user_data)
	{
		dbg (0, "...");

		ayyi_song__print_playlists ();
		am_song__print_tracks      ();
		am_song__channels_print    ();
		am_partmanager__print_list ();

		//FINISH_TEST;
	}

	void __on_song_del(AyyiIdent obj, GError** errror, gpointer user_data)
	{
		//TODO handler order is important. clients may need to run before am_song so that data still exists.

		dbg (0, "OBJECT_SONG");
	}

	AyyiIdent id = {0, -1};
	ayyi_client_add_responder((id.type = AYYI_OBJECT_SONG, id), AYYI_NEW, __on_song_new, NULL);
	ayyi_client_add_responder((id.type = AYYI_OBJECT_SONG, id), AYYI_DEL, __on_song_del, NULL);

	if(app.timeout) g_source_remove (app.timeout);

	printf("waiting for server restart...\n");
}


void
test_never()
{
	module__add_midi_track(0, 0);
	module__add_midi_part(0, 0);
	module__delete_parts_by_audio_file(0, NULL, 0);
	module__delete_audio_part(0, 0, 0);
	module__load_song_with_delay(0, 0, 0);
	module__save_and_load_song_with_delay(0, 0, 0);
	module__delete_all_audio_files(0, 0);
}


